@@ -6,7 +6,7 @@
 ;; Version: 0.2
 ;; Keywords: languages, tools
 ;; URL: https://github.com/wentasah/meson-mode
-;; Package-Requires: ((emacs "24.3"))
+;; Package-Requires: ((emacs "26.1"))
 
 ;; This program is free software; you can redistribute it and/or modify
 ;; it under the terms of the GNU General Public License as published by
@@ -23,11 +23,11 @@
 
 ;;; Commentary:
 
-;; This is a major mode for Meson build system files. Syntax
-;; highlighting works reliably. Indentation works too, but there are
-;; probably cases, where it breaks. Simple completion is supported via
-;; `completion-at-point'. To start completion, use either <C-M-i> or
-;; install completion frameworks such as `company'. To enable
+;; This is a major mode for Meson build system files.  Syntax
+;; highlighting works reliably.  Indentation works too, but there are
+;; probably cases, where it breaks.  Simple completion is supported
+;; via `completion-at-point'.  To start completion, use either C-M-i
+;; or install completion frameworks such as `company'.  To enable
 ;; `company' add the following to your .emacs:
 ;;
 ;;     (add-hook 'meson-mode-hook 'company-mode)
@@ -36,6 +36,8 @@
 ;;; Code:
 
 (require 'compile)
+(require 'seq)
+(eval-when-compile (require 'subr-x))
 
 (defvar meson-mode-syntax-table
   (let ((table (make-syntax-table))
@@ -52,14 +54,15 @@
 		    ?\> "."
 		    ?\= "."
 		    ?\/ "."
-		    ?\| "."
-		    )))
+		    ?\| ".")))
     (while list
       (modify-syntax-entry (pop list) (pop list) table))
     table)
   "Syntax table used while in `meson-mode'.")
 
 (defun meson--max-length (&rest args)
+  "Return maximum length among strings in ARGS.
+If some arguments are numbers, threat them as string lengths."
   (let ((lengths
 	 (mapcar (lambda (x) (if (stringp x) (length x) x)) args)))
     (apply 'max lengths)))
@@ -313,9 +316,7 @@
 	;; array
 	"length"
 	"contains"
-	"get"
-
-	))))
+	"get"))))
 
 (defconst meson--pch-kwargs '("c_pch" "cpp_pch"))
 
@@ -362,8 +363,7 @@
     "objects"
     "override_options"
     "sources"
-    "gnu_symbol_visibility"
-    ))
+    "gnu_symbol_visibility"))
 
 (defconst meson--known-build-target-kwargs
   `(,@meson--buildtarget-kwargs
@@ -454,8 +454,7 @@
 	"version"
 	"private_headers"
 	"cmake_args"
-	"include_type"
-	))
+	"include_type"))
     ("declare_dependency"
      . ("include_directories"
 	"link_with"
@@ -465,8 +464,7 @@
 	"link_args"
 	"link_whole"
 	"version"
-	"variables"
-	))
+	"variables"))
     ("executable"
      . ,meson--known-exe-kwargs)
     ("find_program"
@@ -544,7 +542,7 @@
 
 (defconst meson-syntax-propertize-function
   (syntax-propertize-rules
-   ((rx (or "'''" "'")) (0 (ignore (meson-syntax-stringify))))))
+   ((rx (or "'''" "'")) (0 (ignore (meson--syntax-stringify))))))
 
 (defsubst meson-syntax-count-quotes (&optional point limit)
   "Count number of quotes after point (max is 3).
@@ -558,7 +556,7 @@ and LIMIT is used to limit the scan."
       (setq i (1+ i)))
     i))
 
-(defun meson-syntax-stringify ()
+(defun meson--syntax-stringify ()
   "Put `syntax-table' property correctly on single/triple apostrophes."
   ;; Inspired by python-mode
   (let* ((num-quotes (length (match-string-no-properties 0)))
@@ -593,9 +591,35 @@ and LIMIT is used to limit the scan."
            (put-text-property quote-starting-pos quote-ending-pos
                               'syntax-table (string-to-syntax "|"))))))
 
+(defun meson-function-at-point ()
+  "Return name of the function under point.
+The point can be anywhere within function name or argument list."
+  (save-excursion
+    (skip-syntax-backward "w_")
+    (let ((ppss (syntax-ppss))
+	  (functions (mapcar 'car meson-builtin-functions)))
+      (cond
+       ((or (nth 3 ppss)		; inside string
+	    (nth 4 ppss))		; inside comment
+	(goto-char (nth 8 ppss))	; go to the beginning
+	(meson-function-at-point))
+       ((cl-some (lambda (fname) (when (looking-at fname) fname)) functions))
+       ((and (> (nth 0 ppss) 0)			 ; inside parentheses
+	     (eq (char-after (nth 1 ppss)) ?\()) ; rounded parentheses
+	(goto-char (nth 1 ppss))
+	(cl-some (lambda (fname)
+		   (when (looking-back (concat fname (rx (zero-or-more (any " " "\t"))))
+				       (line-beginning-position))
+		     fname))
+		 functions))
+       ((> (nth 0 ppss) 1)		; inside nested parentheses - other than rounded
+	(goto-char (nth 1 ppss))
+	(meson-function-at-point))))))
+
 ;;; Completion
 
 (defun meson-completion-at-point-function ()
+  "Return possible completion candidates."
   (save-excursion
     (let* ((end (progn (skip-syntax-forward "w_")
 		       (point)))
@@ -639,17 +663,18 @@ and LIMIT is used to limit the scan."
 
 (require 'smie)
 
-(defun meson--comment-bolp (&optional ppss_)
-  "Return non-nil if point is at the beginning of line, ignoring
-comments."
+(defun meson--comment-bolp (&optional syn-ppss)
+  "Return non-nil if point is at the beginning of line, ignoring comments.
+Optional SYN-PPSS is the value returned by `syntax-ppss'."
   (save-excursion
-    (let ((ppss (or ppss_
+    (let ((ppss (or syn-ppss
 		    (syntax-ppss))))
       (when (nth 4 ppss) 		; inside comment
 	(goto-char (nth 8 ppss)))	; go to its beginning
       (smie-rule-bolp))))
 
 (defun meson-smie-forward-token ()
+  "Move forward by one lexer token."
   (let ((token 'unknown))
     (while (eq token 'unknown)
       (let ((ppss (syntax-ppss)))
@@ -700,6 +725,7 @@ comments."
     token))
 
 (defun meson-smie-backward-token ()
+  "Move backward by one lexer token."
   (let ((token 'unknown))
     (while (eq token 'unknown)
       (let ((eopl (max ;; end of previous line (to properly match "eol_cont" below it is actually a character before)
@@ -771,8 +797,7 @@ comments."
 	    ("foreach" foreachblock "endforeach"))
       (foreachblock (id ":" exp "eol" codeblock))
       (ifblock (exp "eol" codeblock)
-	       (exp "eol" codeblock "elif" ifblock)
-	       )
+	       (exp "eol" codeblock "elif" ifblock))
       (exp (exp "," exp)
 	   (id ":" exp)
 	   (exp "+=" exp)
@@ -822,9 +847,7 @@ comments."
       (assoc "==" "!=" "<" "<=" ">" ">=")
       (assoc "+" "-")
       (assoc "*" "/" "%")
-      (assoc ".")
-      )
-    )))
+      (assoc ".")))))
 
 (defgroup meson nil
   "Meson build system mode for Emacs."
@@ -833,9 +856,19 @@ comments."
 
 (defcustom meson-indent-basic 2
   "Indentation offset for meson.build files."
-  :type 'integer)
+  :type 'integer
+  :safe 'integerp)
+
+(defcustom meson-markdown-docs-dir (seq-find 'file-exists-p
+					     '("/usr/share/doc/meson/markdown"
+					       "~/src/meson/docs/markdown"))
+  "Directory containing Meson markdown-formated documentation."
+  :type 'directory)
 
 (defun meson-smie-rules (kind token)
+  "Indentation rules for the SMIE engine.
+See the SMIE documentation for the meaning of KIND and TOKEN
+arguments."
   (pcase (cons kind token)
     (`(:elem . basic) meson-indent-basic)
     (`(:elem . args) (- (save-excursion (beginning-of-line-text) (point)) (point)))
@@ -859,33 +892,85 @@ comments."
     (`(:before . "elif") (smie-rule-parent))
     (_ nil)))
 
-(defun meson-eldoc-documentation-function ()
+;;; Documentation
+
+(defun meson--eldoc-documentation-function ()
   "`eldoc-documentation-function' (which see) for Meson mode."
-  (save-excursion
-    (let* ((end (progn (skip-syntax-forward "w_")
-		       (point)))
-	   (start (progn (skip-syntax-backward "w_")
-			 (point)))
-	   (ppss (syntax-ppss)))
-      (cond
-       ((or (nth 3 ppss)		; inside string
-	    (nth 4 ppss))		; inside comment
-	nil) ; nothing to complete
-       ((cl-some (lambda (func)
-		   (when  (looking-at (car func))
-		     (plist-get (cdr func) :doc)))
-		 meson-builtin-functions))))))
+  (if-let* ((fname (meson-function-at-point))
+	    (fspec (alist-get fname meson-builtin-functions)))
+      (plist-get fspec :doc)))
+
+(defun meson--find-reference-manual ()
+  "Return the absolute filename of the Meson reference manual or nil."
+  (let ((default-directory meson-markdown-docs-dir))
+    (when-let (manual (seq-find #'file-exists-p '("Reference-manual.md"
+                                                  "Reference-manual.md.gz")))
+      (expand-file-name manual))))
+
+(defun meson--make-lookup-regexp (identifier)
+  "Make regexp for looking up IDENTIFIER in the Meson reference manual."
+  ;; In Emacs 27 this could be simplified to (rx ... (literal identifier) ...).
+  (rx-to-string
+   `(seq bol (or (seq (+ "#") " " (? "`") ,identifier (or "(" "`" eol))
+                 (seq (* blank) "- `" ,identifier "(")))))
+
+(defun meson--search-in-reference-manual (identifier)
+  "Search for the function or object IDENTIFIER in the current buffer.
+The current buffer is assumed to contain the Meson reference manual.
+Return either `line-beginning-position' of the matching line or nil."
+  (goto-char (point-min))
+  (and (re-search-forward (meson--make-lookup-regexp identifier) nil t)
+       (line-beginning-position)))
+
+(defun meson-lookup-doc (identifier)
+  "Open Meson reference manual and find IDENTIFIER.
+Return the buffer containing the reference manual.
+IDENTIFIER is the name of a Meson function or object as a string.
+Signal a `user-error' if the manual could not be found
+or does not contain IDENTIFIER."
+  (interactive (list (or (thing-at-point 'symbol)
+                         (meson-function-at-point)
+                         (user-error "No identifier at point"))))
+  (let ((buf (find-file-noselect
+              (or (meson--find-reference-manual)
+                  (user-error "Meson reference manual not found")))))
+    (with-current-buffer buf
+      ;; Set up buffer only once after creation.
+      (unless (string= (buffer-name) "*Meson Reference Manual*")
+        (rename-buffer "*Meson Reference Manual*" 'unique)
+        (read-only-mode)
+        (when (and (require 'markdown-mode nil t)
+                   (fboundp 'markdown-view-mode)
+                   (not (eq major-mode 'markdown-view-mode)))
+          (markdown-view-mode))
+        (local-set-key (kbd "q") 'bury-buffer)
+        (when (bound-and-true-p evil-mode)
+          (evil-local-set-key 'normal (kbd "q") 'bury-buffer)))
+      (goto-char
+       (or (meson--search-in-reference-manual identifier)
+           (user-error "%s not found in Meson reference manual" identifier))))
+    (switch-to-buffer buf)
+    (recenter 0)
+    buf))
+
+(defalias 'meson-lookup-doc-at-point (symbol-function 'meson-lookup-doc))
 
 ;;; Mode definition
 
+(defvar meson-mode-map
+       (let ((map (make-sparse-keymap)))
+         (define-key map [f1] 'meson-lookup-doc)
+         map)
+       "Keymap for `meson-mode'.")
+
 ;;;###autoload
 (define-derived-mode meson-mode prog-mode "Meson"
-  "Major mode for editing Meson build system files."
+  "Major mode for editing Meson build system files.
+\\{meson-mode-map}"
   :abbrev-table nil
   (setq font-lock-defaults
 	'(meson-mode-font-lock-keywords
-	  nil nil nil nil
-	  ))
+	  nil nil nil nil))
 
   (set (make-local-variable 'syntax-propertize-function)
        meson-syntax-propertize-function)
@@ -898,8 +983,7 @@ comments."
 	      :forward-token #'meson-smie-forward-token
 	      :backward-token #'meson-smie-backward-token)
   (add-function :before-until (local 'eldoc-documentation-function)
-                #'meson-eldoc-documentation-function)
-  )
+                #'meson--eldoc-documentation-function))
 
 ;;;###autoload
 (progn
